# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import time

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobQueued
from colcon_core.event.output import StderrLine
from colcon_core.event.test import TestFailure
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_handler import format_duration
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import SIGINT_RESULT


class SummaryHandler(EventHandlerExtensionPoint):
    """
    Output summary of all tasks.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.output.StderrLine`
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.job.JobQueued`
    - :py:class:`colcon_core.event.test.TestFailure`
    - :py:class:`colcon_core.event_reactor.EventReactorShutdown`
    """

    # the priority should be lower than other extensions
    # in order to not output the summary before they are finished
    PRIORITY = 50

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._queued = set()
        self._with_stderr = set()
        self._with_test_failures = set()
        self._ended = set()
        self._failed = set()
        self._interrupted = set()
        self._start_time = time.monotonic()

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, JobQueued):
            job = event[1]
            self._queued.add(job)
        elif isinstance(data, StderrLine):
            job = event[1]
            self._with_stderr.add(job)
        elif isinstance(data, TestFailure):
            job = event[1]
            self._with_test_failures.add(job)
        elif isinstance(data, JobEnded):
            job = event[1]
            self._ended.add(job)
            if data.rc == SIGINT_RESULT:
                self._interrupted.add(job)
            elif data.rc:
                self._failed.add(job)
        elif isinstance(data, EventReactorShutdown):
            self._print_summary()

    def _print_summary(self):
        # separate the summary from the previous output
        print()

        duration = time.monotonic() - self._start_time
        duration_string = format_duration(duration)

        blocked_jobs = self._queued - self._ended

        # convert jobs to package names
        ended = {j.task_context.pkg for j in self._ended}
        interrupted = {j.task_context.pkg for j in self._interrupted}
        failed = {j.task_context.pkg for j in self._failed}
        with_stderr = {j.task_context.pkg for j in self._with_stderr}
        with_test_failures = {
            j.task_context.pkg for j in self._with_test_failures}
        blocked = {j.task_context.pkg for j in blocked_jobs}

        # packages with successful jobs and blocked jobs are "interrupted"
        interrupted |= ended & blocked - failed

        # truly "blocked" packages have no jobs attempted
        blocked -= ended

        count, job_type, _ = _msg_arguments(ended - interrupted - failed)
        print('Summary: {count} {job_type} finished '
              '[{duration_string}]'.format_map(locals()))

        if failed:
            count, job_type, names = _msg_arguments(failed)
            print('  {count} {job_type} failed: {names}'
                  .format_map(locals()))

        if interrupted:
            count, job_type, names = _msg_arguments(interrupted)
            print('  {count} {job_type} aborted: {names}'
                  .format_map(locals()))

        if with_stderr:
            count, job_type, names = _msg_arguments(with_stderr)
            print(
                '  {count} {job_type} had stderr output: {names}'
                .format_map(locals()))

        if with_test_failures:
            count, job_type, names = _msg_arguments(with_test_failures)
            print(
                '  {count} {job_type} had test failures: {names}'
                .format_map(locals()))

        if blocked:
            count = len(blocked)
            job_type = get_job_type_word_form(count)
            print(
                '  {count} {job_type} not processed'
                .format_map(locals()))


def _msg_arguments(packages):
    return (
        len(packages),
        get_job_type_word_form(len(packages)),
        ' '.join(sorted(p.name for p in packages)),
    )


def get_job_type_word_form(job_count):
    """
    Get the singular / plural word form of a job type.

    While this function only returns "package" or "packages" it allows external
    code to replace the function with custom logic.

    :param job_count: The number of jobs
    :rtype: str
    """
    return 'package' if job_count == 1 else 'packages'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import copy
import errno
import locale
import os
import time

from colcon_core.event.command import Command
from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobStarted
from colcon_core.event.output import StderrLine
from colcon_core.event.output import StdoutLine
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_handler.console_direct import ConsoleDirectEventHandler
from colcon_core.location import create_log_path
from colcon_core.location import get_log_path
from colcon_core.plugin_system import satisfies_version

# when this extension is present it disables the direct console output
ConsoleDirectEventHandler.ENABLED_BY_DEFAULT = False

COMMAND_LOG_FILENAME = 'command.log'
STDOUT_LOG_FILENAME = 'stdout.log'
STDERR_LOG_FILENAME = 'stderr.log'
STDOUT_STDERR_LOG_FILENAME = 'stdout_stderr.log'
ALL_STREAMS_LOG_FILENAME = 'streams.log'

all_log_filenames = [
    COMMAND_LOG_FILENAME,
    STDOUT_LOG_FILENAME,
    STDERR_LOG_FILENAME,
    STDOUT_STDERR_LOG_FILENAME,
    ALL_STREAMS_LOG_FILENAME,
]


class LogEventHandler(EventHandlerExtensionPoint):
    """
    Output task specific log files.

    The following log files are created in the log directory of the specific
    task: `command.log`, `stdout.log`, `stderr.log`.
    Additionally the log file `streams.log` is created with the combined
    content.

    The presence of this extension disables the default console output in order
    to keep the amount of output readable.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.command.Command`
    - :py:class:`colcon_core.event.job.JobStarted`
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.output.StdoutLine`
    - :py:class:`colcon_core.event.output.StderrLine`
    """

    # the priority should be higher than the default priority
    # in order to write the information to the log files before
    PRIORITY = 150

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._jobs = set()
        self._file_handles = {}
        self._start_times = {}

    def __call__(self, event):  # noqa: D102
        data = event[0]
        job = event[1]

        if isinstance(data, JobStarted):
            self._start_times[job] = time.monotonic()

        if isinstance(data, JobEnded):
            # Skip if the log path is /dev/null
            # => is unable to get directory
            base_path = get_log_directory(job)
            if base_path is None:
                return

            for filename in all_log_filenames:
                path = base_path / filename
                if path in self._file_handles:
                    self._file_handles[path].close()
            return

        filenames = copy.copy(all_log_filenames)
        if not isinstance(data, Command):
            filenames.remove(COMMAND_LOG_FILENAME)
        if not isinstance(data, StdoutLine):
            filenames.remove(STDOUT_LOG_FILENAME)
        if not isinstance(data, StderrLine):
            filenames.remove(STDERR_LOG_FILENAME)
        if (
            not isinstance(data, StdoutLine) and
            not isinstance(data, StderrLine)
        ):
            filenames.remove(STDOUT_STDERR_LOG_FILENAME)
        if len(filenames) <= 1:
            # skip if event is neither of the known events
            return

        if not self._init_logs(job):
            return

        if isinstance(data, Command):
            line = data.to_string() + '\n'
        else:
            line = data.line

        if not isinstance(line, bytes):
            # use the same encoding as the default for the opened file
            line = line.encode(encoding=locale.getpreferredencoding(False))

        # Get base path, and return if fails
        base_path = get_log_directory(job)
        if base_path is None:
            return

        for filename in filenames:
            h = self._file_handles[base_path / filename]

            if filename == ALL_STREAMS_LOG_FILENAME:
                # prefix line with relative time
                relative_time = time.monotonic() - self._start_times[job]
                # use the same encoding as the default for the opened file
                prefix = ('[%.3fs] ' % relative_time).encode(
                    encoding=locale.getpreferredencoding(False))
                h.write(prefix)

            h.write(line)
            try:
                h.flush()
            except OSError as e:
                if e.errno == errno.ENOSPC:
                    # for known error code suppress the stacktrace
                    # and only show the exception message
                    raise RuntimeError(
                        str(e) + ' [{h.name}]'.format_map(locals()))
                raise

    def _init_logs(self, job):
        # only create logs once per task
        if job in self._jobs:
            return True

        # Get directory
        # If this fails, the log directory is /dev/null
        # so function must return with negative result
        base_path = get_log_directory(job)
        if base_path is None:
            return False

        self._jobs.add(job)

        create_log_path(self.context.args.verb_name)

        os.makedirs(str(base_path), exist_ok=True)
        for filename in all_log_filenames:
            path = base_path / filename
            self._file_handles[path] = path.open(mode='wb')
        return True


def get_log_directory(job):
    """
    Get the log directory for a specific job.

    :param job: The job
    :rtype: Path
    """
    log_path = get_log_path()
    if log_path is None:
        return None
    return log_path / job.identifier



# Copyright 2026 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from unittest.mock import Mock

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobQueued
from colcon_core.event.output import StderrLine
from colcon_core.event.test import TestFailure as _TestFailure
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.executor import Job
from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.subprocess import SIGINT_RESULT
from colcon_core.task import TaskContext
from colcon_output.event_handler.summary import SummaryHandler
import pytest


@pytest.fixture
def summary_handler():
    return SummaryHandler()


def _create_mock_job(name, pkg=None):
    if pkg is None:
        pkg = PackageDescriptor(Path('/path/to/' + name))
        pkg.name = name

    context = TaskContext(pkg=pkg, args=None, dependencies=None)

    task = Mock()
    task.context = context

    job = Job(
        identifier=name, dependencies=set(), task=task,
        task_context=context)
    return job


def test_successful_job(summary_handler, capsys):
    job = _create_mock_job('pkg_a')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    summary_handler((JobEnded('pkg_a', 0), job))
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '1 package finished' in captured.out


def test_failed_job(summary_handler, capsys):
    job = _create_mock_job('pkg_fail')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    summary_handler((JobEnded('pkg_fail', 1), job))
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '0 packages finished' in captured.out
    assert '1 package failed: pkg_fail' in captured.out


def test_interrupted_job(summary_handler, capsys):
    job = _create_mock_job('pkg_abort')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    summary_handler((JobEnded('pkg_abort', SIGINT_RESULT), job))
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '0 packages finished' in captured.out
    assert '1 package aborted: pkg_abort' in captured.out


def test_stderr_job(summary_handler, capsys):
    job = _create_mock_job('pkg_stderr')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    summary_handler((StderrLine(b'error'), job))
    summary_handler((JobEnded('pkg_stderr', 0), job))
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '1 package finished' in captured.out
    assert '1 package had stderr output: pkg_stderr' in captured.out


def test_test_failure_job(summary_handler, capsys):
    job = _create_mock_job('pkg_test_fail')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    summary_handler((_TestFailure('pkg_test_fail'), job))
    summary_handler((JobEnded('pkg_test_fail', 0), job))
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '1 package finished' in captured.out
    assert '1 package had test failures: pkg_test_fail' in captured.out


def test_blocked_job(summary_handler, capsys):
    job = _create_mock_job('pkg_blocked')
    summary_handler((JobQueued(job.task_context.pkg.name), job))
    # Never ends
    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()
    assert '0 packages finished' in captured.out
    assert '1 package not processed' in captured.out


def test_multi_job_interrupted(summary_handler, capsys):
    # One job succeeds, one job blocked -> the package is considered aborted
    shared_pkg = PackageDescriptor(Path('/path/to/pkg_multi'))
    shared_pkg.name = 'pkg_multi'
    shared_pkg.type = 'mock'

    job1 = _create_mock_job('pkg_multi', pkg=shared_pkg)
    job2 = _create_mock_job('pkg_multi', pkg=shared_pkg)

    summary_handler((JobQueued(job1.task.context.pkg.name), job1))
    summary_handler((JobQueued(job2.task.context.pkg.name), job2))

    summary_handler((JobEnded('pkg_multi', 0), job1))
    # job2 never ends

    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()

    # "interrupted |= ended & blocked - failed" means it's aborted,
    # not finished, not blocked
    assert '0 packages finished' in captured.out
    assert '1 package aborted: pkg_multi' in captured.out


def test_multi_job_failed_early(summary_handler, capsys):
    shared_pkg = PackageDescriptor(Path('/path/to/pkg_multi_fail_early'))
    shared_pkg.name = 'pkg_multi_fail_early'

    job1 = _create_mock_job('pkg_multi_fail_early', pkg=shared_pkg)
    job2 = _create_mock_job('pkg_multi_fail_early', pkg=shared_pkg)

    summary_handler((JobQueued(job1.task.context.pkg.name), job1))
    summary_handler((JobQueued(job2.task.context.pkg.name), job2))

    summary_handler((JobEnded('pkg_multi_fail_early', 1), job1))
    # job2 never ends

    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()

    assert '0 packages finished' in captured.out
    assert '1 package failed: pkg_multi_fail_early' in captured.out


def test_multi_job_failed_late(summary_handler, capsys):
    shared_pkg = PackageDescriptor(Path('/path/to/pkg_multi_fail_late'))
    shared_pkg.name = 'pkg_multi_fail_late'

    job1 = _create_mock_job('pkg_multi_fail_late', pkg=shared_pkg)
    job2 = _create_mock_job('pkg_multi_fail_late', pkg=shared_pkg)

    summary_handler((JobQueued(job1.task.context.pkg.name), job1))
    summary_handler((JobQueued(job2.task.context.pkg.name), job2))

    summary_handler((JobEnded('pkg_multi_fail_late', 0), job1))
    summary_handler((JobEnded('pkg_multi_fail_late', 1), job2))

    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()

    assert '0 packages finished' in captured.out
    assert '1 package failed: pkg_multi_fail_late' in captured.out


def test_multi_job_stderr(summary_handler, capsys):
    shared_pkg = PackageDescriptor(Path('/path/to/pkg_multi_stderr'))
    shared_pkg.name = 'pkg_multi_stderr'

    job1 = _create_mock_job('pkg_multi_stderr', pkg=shared_pkg)
    job2 = _create_mock_job('pkg_multi_stderr', pkg=shared_pkg)

    summary_handler((JobQueued(job1.task.context.pkg.name), job1))
    summary_handler((JobQueued(job2.task.context.pkg.name), job2))

    summary_handler((StderrLine(b'error'), job1))
    summary_handler((JobEnded('pkg_multi_stderr', 0), job1))
    summary_handler((JobEnded('pkg_multi_stderr', 0), job2))

    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()

    assert '1 package finished' in captured.out
    assert '1 package had stderr output: pkg_multi_stderr' in captured.out


def test_multi_job_test_failure(summary_handler, capsys):
    shared_pkg = PackageDescriptor(Path('/path/to/pkg_multi_test_fail'))
    shared_pkg.name = 'pkg_multi_test_fail'

    job1 = _create_mock_job('pkg_multi_test_fail', pkg=shared_pkg)
    job2 = _create_mock_job('pkg_multi_test_fail', pkg=shared_pkg)

    summary_handler((JobQueued(job1.task.context.pkg.name), job1))
    summary_handler((JobQueued(job2.task.context.pkg.name), job2))

    summary_handler((JobEnded('pkg_multi_test_fail', 0), job1))
    summary_handler((_TestFailure('pkg_multi_test_fail'), job2))
    summary_handler((JobEnded('pkg_multi_test_fail', 0), job2))

    summary_handler((EventReactorShutdown(), None))
    captured = capsys.readouterr()

    assert '1 package finished' in captured.out
    assert '1 package had test failures: pkg_multi_test_fail' in captured.out



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import errno
import time

from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.location import create_log_path
from colcon_core.location import get_log_path
from colcon_core.plugin_system import satisfies_version


class EventLogEventHandler(EventHandlerExtensionPoint):
    """
    Log all events to a global log file.

    The log file `events.log` is created in the log directory.
    """

    FILENAME = 'events.log'

    # the priority should be higher than the default priority
    # in order to write the events to the log file first
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._file_handle = None
        self._start_time = None

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if not self._init_log():
            return

        context = str(event[1]) if event[1] is not None else '-'
        try:
            members = data.__dict__
        except AttributeError:
            members = {s: getattr(data, s) for s in data.__slots__}
        self._file_handle.write(
            '[%f] (%s) %s: %s\n' % (
                self._get_relative_time(), context,
                data.__class__.__name__, members))
        try:
            self._file_handle.flush()
        except OSError as e:
            if e.errno == errno.ENOSPC:
                # for known error code suppress the stacktrace
                # and only show the exception message
                raise RuntimeError(
                    str(e) +
                    ' [{self._file_handle.name}]'.format_map(locals()))
            raise

        if isinstance(data, EventReactorShutdown):
            self._file_handle.close()
            self._file_handle = None

    def _init_log(self):
        # only create log once
        if self._file_handle is not None:
            return True

        log_path = get_log_path()
        if log_path is None:
            return False

        create_log_path(self.context.args.verb_name)
        path = log_path / EventLogEventHandler.FILENAME
        self._file_handle = path.open(mode='w')
        return True

    def _get_relative_time(self):
        now = time.monotonic()
        if self._start_time is None:
            self._start_time = now
            return 0
        return now - self._start_time
